"""
FileUpload Class
================
"""
import os
from warnings import warn
from flask import send_from_directory, Flask, request
from werkzeug.utils import secure_filename
from typing import Any, List, Dict, Union

from ._config import Config
from .model import Model
from .column import Column
from .file_utils import FileUtils
from ._model_utils import _ModelUtils


class FileUpload:
    """
    :param app: Flask application instance
    """

    #: The Flask application instance: ``app = Flask(__name__)``.
    #: We can either pass the instance to ``FileUpload(app)`` or
    #: to the ``init_app(app)`` method::
    #:
    #:    app = Flask(__name__)
    #:
    #:    db = SQLAlchemy()
    #:    file_upload = FileUpload()
    #:
    #:    # An example using the Flask factory pattern
    #:    def create_app():
    #:        db.init_app(app)
    #:        file_upload.init_app(app)
    #:
    #:    # Or we can pass the Flask app instance directly:
    #:    db = SQLAlchemy(app)
    #:    file_upload = FileUpload(app)
    app: Flask = None

    #: The configuration class used for this library.
    #: See :class:`~flask_file_upload._config` for more information.
    config: Config = Config()

    #: All the file related model attributes & values are stored
    #: here as a list of dicts.
    file_data: List[Dict[str, str]] = []

    #: A record of the original filenames used when saving files
    #: to the server.
    files: Any = []

    #: A class containing utility methods for working with files.
    #: See :class:`flask_file_upload.file_utils` for more information.
    file_utils: FileUtils = None

    def __init__(self, app=None):
        """
        The Flask application instance: ``app = Flask(__name__)``.
        :param app: Flask application instance
        """
        self.Model = Model
        self.Column = Column
        if app:
            self.init_app(app)

    def delete_files(self, model: Any, db=None, **kwargs) -> Union[Any, None]:
        """
        Public method for removing stored files from the server & database.
        This method will remove all files passed to the kwarg ``files`` list.
        It will also update the passed in SqlAlchemy ``model`` object & return
        the updated model object if ``db`` is None.

        If the ``db`` arg is passed in then the session is updated & session commited &
        this method return value is void::

            # Example using a SqlAlchemy model with an appended
            # method that fetches a single `blog`
            blogModel = BlogModel()
            blog_results = blogModel.get_one()

            # We pass the blog
            blog = file_upload.delete_files(blog_result, files=["my_video"])

            # As the `db` arg has not been passed to this method,
            # the changes would need persisting to the database:
            db.session.add(blog)
            db.session.commit()

            # If `db` is passed to this method then the updates are persisted.
            # to the session. And therefore the session has been commited &
            # no blog is returned.
            file_upload.delete_files(blog_result, db, files=["my_video"])


        :param model: Instance of a SqlAlchemy Model
        :param db: Either an instance of Flask-SqlAlchemy ``SqlAlchmey`` class or
               SqlAlchmey's ``Session`` object.
        :key files: A list of the file names declared on your model.
        :return: SqlAlchemy model object
        """
        try:
            files: List[str] = kwargs["files"]
        except KeyError:
            warn("'files' is a Required Argument")
            return None

        self.file_utils = FileUtils(model, self.config)

        for f in files:
            file_type = _ModelUtils.get_by_postfix(model, f, 'file_type')
            file_path = f"{self.file_utils.get_stream_path(model.id)}/{f}.{file_type}"
            os.remove(f"{file_path}")

        for f_name in files:
            for postfix in _ModelUtils.keys:
                setattr(model, _ModelUtils.add_postfix(f_name, postfix), None)
        if db:
            db.session.add(model)
            db.session.commit()
        else:
            return model

    def _check_attrs(self, model: Any, attr: str):
        """
        Before we can set the attribute on the Model we check
        that this attributes exists so it matches with the
        db's table columns.
        :param model:
        :param attr:
        :return:
        """
        if not hasattr(model, attr):
            raise AttributeError(
                f"Flask-File-Upload: Attribute {attr} does not exist on your model, "
                "please check your files has been declared correctly on your model. "
                "See https://github.com/joegasewicz/Flask-File-Upload"
            )

    def _create_file_dict(self, file, attr_name: str):
        """
        :param file:
        :param attr_name:
        :return:
        """
        if file.filename != "" and file and FileUtils.allowed_file(file.filename, self.config):
            filename = secure_filename(file.filename)
            filename_key = attr_name
            mime_type = file.content_type
            file_type = file.filename.split(".")[1]
            return {
                f"{filename_key}__{_ModelUtils.keys[0]}": filename,
                f"{filename_key}__{_ModelUtils.keys[1]}": mime_type,
                f"{filename_key}__{_ModelUtils.keys[2]}": file_type,
            }
        else:
            warn("Flask-File-Upload: No files were saved")
            return {}

    def get_file_url(self, model: Any, **kwargs) -> str:
        """
        Returns the url path to a file on your server.
        To access the file, it must reference the attribute name
        defined on your SqlAlchemy model. For example::

            @file_upload.Model
            class ModelTest(db.Model):

                my_video = file_upload.Column(db)

        To return the url of the above file, pass in attribuute name
        to the ``filename`` kwarg. Example::

            file_upload.get_file_url(blog_post, filename="my_video")

        :param model:
        :param kwargs:
        :return:
        """
        try:
            filename = kwargs["filename"]
            self.file_utils = FileUtils(model, self.config)
            file_path = self.file_utils.get_file_path(model.id, filename)
            file_type = _ModelUtils.get_by_postfix(model, filename, "file_type")
            return f"{request.url}{file_path}.{file_type}"
        except AttributeError:
            AttributeError("[FLASK_FILE_UPLOAD] You must declare a filename kwarg")

    def init_app(self, app) -> None:
        """
        If you are using the Flask factory pattern, normally you
        will, by convention, create a method called ``create_app``.
        At runtime, you may then set configuration values before initiating
        FileUpload etc. Example::

            app = Flask(__name__)
            file_upload = FileUpload()

            def create_app():
                file_upload.init_app(app)

        :param app: The Flask application instance: ``app = Flask(__name__)``.
        :return: None
        """
        self.app = app
        self.config.init_config(app)

    def save_files(self, model, **kwargs) -> Any:
        """
        The file(s) must reference the attribute name(s)
        defined in your SqlAlchemy model. For example::

            @file_upload.Model
            class ModelTest(db.Model):

                my_video = file_upload.Column(db)
                placeholder_img = file_upload.Column(db)

        This example demonstrates creating a new row in your database
        using a SqlAlchemy model which is is then pass as the first
        argument to ``file_upload.save_files``. Normally, you will
        access your files from Flask's ``request`` object::

            from Flask import request

            my_video = request.files['my_video']
            placeholder_img = request.files['placeholder_img']

        Then, we need to pass to the kwarg ``files`` a dict of keys that
        reference the attribute name(s) defined in your SqlAlchemy
        model & values that are your files.::

            blog_post = BlogPostModel(title="Hello World Today")

            file_upload.save_files(blog_post, files={
                "my_video": my_video,
                "placeholder_img": placeholder_img,
            })

        :param model: The SqlAlchemy model instance
        :key filename: The attribute name(s) defined in your SqlAlchemy
            model
        :return: The updated SqlAlchemy model instance
        """
        # Warning: These methods need to set members on the Model class
        # before we instantiate FileUtils()
        self._set_file_data(**kwargs)
        self._set_model_attrs(model)

        self.file_utils = FileUtils(model, self.config)

        # Save files to dirs
        self._save_files_to_dir(model)

        return model

    def _save_files_to_dir(self, model: Any) -> None:
        """
        :param model:
        :return None:
        """
        for f in self.files:
            id_val = _ModelUtils.get_id_value(model)
            self.file_utils.save_file(f, id_val)

    def _set_file_data(self, **file_data) -> List[Dict[str, str]]:
        """
        Adds items to files & file_data
        :key files: Dict[str: Any] Key is the filename & Value
        is the file.
        :return:  List[Dict[str, str]]
        """
        for k, v in file_data.get("files").items():
            self.files.append(v)
            self.file_data.append(self._create_file_dict(v, k))
        return self.file_data

    def _set_model_attrs(self, model: Any) -> None:
        """
        :param model:
        :return: None
        """
        for d in self.file_data:
            for k, v in d.items():
                self._check_attrs(model, k)
                setattr(model, k, v)

    def stream_file(self, model, **kwargs) -> Any:
        """
        Streams a file from the directory defined by
        the Flask's ``UPLOAD_FOLDER``, in your Flask app's
        configuration settings. To access the file,
        it must reference the attribute name defined on
        your SqlAlchemy model. For example::

            @file_upload.Model
            class ModelTest(db.Model):

                my_video = file_upload.Column(db)

        The required steps to stream files using ``file_upload.stream_file`` are:

        1. Fetch the required row from your database using SqlAlchemy::

            blog = BlogModel.query.filter_by(id=id).first()

        2. Pass the SqlAlchemy instance to ``file_upload.stream_file`` &
           reference the attribute name defined on your SqlAlchemy model::

            file_upload.stream_file(blogs, filename="my_video")

        :param model: SqlAlchemy model instance.
        :key filename: The attribute name defined on your SqlAlchemy model
        :return: Updated SqlAlchemy model instance or None.
        """
        try:
            filename = kwargs['filename']
        except KeyError:
            warn("'files' is a Required Argument")
            return None

        file_type = _ModelUtils.get_by_postfix(model, filename, _ModelUtils.keys[1])

        self.file_utils = FileUtils(model, self.config)

        return send_from_directory(
            self.file_utils.get_stream_path(model.id),
            f"{filename}.{file_type}",
            conditional=True,
        )

    def update_files(self, model: Any, db=None, **kwargs):
        """
        First reference the attribute name defined on your
        SqlAlchemy model. For example::

            @file_upload.Model
            class ModelTest(db.Model):

                my_video = file_upload.Column(db)

        Pass the model instance as a first argument to ``file_upload.update_files``.
        The kwarg ``files`` requires the file(s) returned from Flask's request body::

            from Flask import request

            my_video = request.files['my_video']
            placeholder_img = request.files['placeholder_img']

        Then, we need to pass to the kwarg ``files`` a dict of keys that
        reference the attribute name(s) defined in your SqlAlchemy
        model & values that are your files.::


            blog_post = BlogPostModel(title="Hello World Today")
            blog_post = file_upload.update_files(blog_post, files={
                "my_video": new_my_video,
                "placeholder_img": new_placeholder_img,
            })

        :param model: SqlAlchemy model instance.
        :param db: Default is None which is not Recommended. If db is None,
            then only the files on the server are removed & the model is updated with
            each attribute set to None but the session is not commited (This could cause
            your database & files on server to be out of sync if you fail to commit
            the session.
        :key files Dict[str, Any]: A dict with the key representing the model attr
             name & file as value.
        :return Any: Returns the model back
        """
        try:
            files = kwargs["files"]
        except KeyError:
            warn("'files' is a Required Argument")
            return None

        original_file_names = []

        for f in files:
            value = _ModelUtils.get_by_postfix(model, f, "file_name")
            original_file_names.append(value)

        # Set file_data
        self._set_file_data(**kwargs)
        self._set_model_attrs(model)

        self.file_utils = FileUtils(model, self.config)

        # Save files to dirs
        self._save_files_to_dir(model)

        # remove original files from directory
        for f in original_file_names:
            os.remove(f"{self.file_utils.get_stream_path(model.id)}/{f}")

        # if a db arg is provided then commit changes to db
        if db:
            db.session.add(model)
            db.session.commit()
            return None
        else:
            return model



"""
    Behaviours required by Model class _so we can keep SqlAlchemy Model
    free of methods & other members.
"""
from typing import List, Any, Dict, Tuple, ClassVar, Callable

from .column import Column


class _ModelUtils:

    keys: Tuple[str] = ("file_name", "file_type", "mime_type")

    @staticmethod
    def create_keys(keys: Tuple[str], filename: str, fn: Callable = None) -> Dict[str, None]:
        """
        Adds the SqlAlchemy Column object with key & name kwargs defined to the returned dict
        :param keys:
        :param filename:
        :param fn: Generates a SqlAlchemy Column with key & name kwargs set
        :return:
        """
        col_dict = {}
        for k in keys:
            key = f"{filename}__{k}"
            col_dict[key] = fn(key, key)
        return col_dict

    @staticmethod
    def get_primary_key(model):
        """
        This will always target the first primary key in
        the list (in case there are multiple being used)
        :param model:
        :return str:
        """
        try:
            return model.__mapper__.primary_key[0].name
        except AttributeError as err:
            raise AttributeError("[FLASK_FILE_UPLOADS_ERROR] You must pass a model instance"
                                 f"to the save_file method. Full error: {err}"
                                 )

    @staticmethod
    def get_table_name(model: Any) -> str:
        """
        Set on class initiation
        :return: None
        """
        return model.__tablename__

    @staticmethod
    def get_id_value(model) -> int:
        """
        :param model:
        :return:
        """
        return getattr(model, _ModelUtils.get_primary_key(model), None)

    @staticmethod
    def columns_dict(file_name: str, db) -> Dict[str, Any]:
        """
        We must define the SqlAlchemy Column object with key & name kwargs
        otherwise sqlAlchemy will define these incorrectly if they are set to None
        :param file_name:
        :param db:
        :return Dict[str, Any]:
        """
        def create_col(key, name): return db.Column(db.String,  key=key, name=name)
        return _ModelUtils.create_keys(
            _ModelUtils.keys,
            file_name,
            create_col
        )

    @staticmethod
    def set_columns(wrapped: ClassVar, new_cols: Tuple[Dict[str, Any]]) -> None:
        """
        Sets related file data to a SqlAlchemy Model
        :return:
        """
        for col_dict in new_cols:
            for k, v in col_dict.items():
                setattr(wrapped, k, v)

    @staticmethod
    def remove_unused_cols(wrapped: ClassVar,  filenames: Tuple[str]) -> None:
        """
        Removes the original named attributes
        (this could be a good place to store
         metadata in a dict for example...)
        :return:
        """
        for col_name in filenames:
            delattr(wrapped, col_name)

    @staticmethod
    def get_attr_from_model(wrapped: ClassVar, new_cols: List, file_names: List) -> Any:
        """
        Adds values to new_cols & file_names so as not to
        change the size of the dict at runtime
        :return None:
        """
        for attr, value in wrapped.__dict__.items():
            if isinstance(value, Column):
                new_cols.append(_ModelUtils.columns_dict(attr, value.db))
                file_names.append(str(attr))
        return new_cols, file_names

    @staticmethod
    def add_postfix(filename: str, postfix: str) -> str:
        """
        :param filename:
        :param postfix:
        :return str:
        """
        return f"{filename}__{postfix}"

    @staticmethod
    def get_by_postfix(model: ClassVar, filename: str, postfix: str) -> str:
        """
        :param model:
        :param filename:
        :param postfix:
        :return str:
        """
        return getattr(model, _ModelUtils.add_postfix(filename, postfix))



"""
   Public SQLAlchemy class decorator.
"""
from functools import update_wrapper


from ._model_utils import _ModelUtils


class Model:
    """
    Flask-File-Upload (FFU) setup requires each SQLAlchemy model that wants to use
    FFU library to be decorated with ``@file_upload.Model``.This will enable FFU
    to update your database with the extra columns required to store
    files in your database.
    Declare your attributes as normal but assign a value of
    ``file_upload.Column`` & pass the SQLAlchemy ``db`` instance:
    ``file_upload.Column(db)``. This is easy if you are using Flask-SQLAlchemy::

        from flask_SQLAlchemy import SQLAlchemy

        db = SQLAlchemy()

    Full example::

       from my_app import db, file_upload

       @file_upload.Model
       class blogModel(db.Model):
           __tablename__ = "blogs"
           id = db.Column(db.Integer, primary_key=True)

           # Your files -  Notice how we pass in the SQLAlchemy instance
           # or `db` to the `file_uploads.Column` class:

           my_placeholder = file_upload.Column(db)
           my_video = file_upload.Column(db)

    """

    def __init__(self, _class):
        """
        We set all our attributes in this __init__ method so that when SQLAlchemy's
        `create_all` method is evoked the model's attributes are set.
        :param _class:
        """
        update_wrapper(self, _class)
        super(Model, self).__init__()

        self._class = _class

        new_cols = []
        filenames = []

        new_cols_list, filenames_list = _ModelUtils.get_attr_from_model(self._class, new_cols, filenames)
        # Add new attributes to the SQLAlchemy model
        _ModelUtils.set_columns(self._class, new_cols_list)
        # The original model's attributes set by the user for files get removed here
        _ModelUtils.remove_unused_cols(self._class, filenames_list)

    def __call__(self, *args, **kwargs):
        """
        When the wrapped class is instantiated , this method will be called
        with the wrapped constructor kwargs passed in - in this case it would
        be the user's model attribute values
        :param args:
        :param kwargs:
        :return:
        """
        return self._class(*args, **kwargs)






from typing import List
from warnings import warn


class Config:

    upload_folder: str = ""

    server_name: str = ""

    allowed_extensions: List = []

    max_content_length: int = 0

    def init_config(self, app):
        self.server_name = app.config["SERVER_NAME"]
        try:
            self.upload_folder = app.config["UPLOAD_FOLDER"]
        except KeyError as _:
            raise KeyError("Flask-File-Uploads: UPLOAD_FOLDER must be set")
        try:
            self.allowed_extensions = app.config["ALLOWED_EXTENSIONS"]
        except KeyError as _:
            self.allowed_extensions = ["jpg", "png", "mov", "mp4", "mpg"]
            warn("Flask-File-Uploads: ALLOWED_EXTENSIONS is not set."
                 f"Defaulting to: {self.allowed_extensions}")
        self.max_content_length = app.config.get("MAX_CONTENT_LENGTH")



class Column:
    """
    The Column class is used to define the file attributes on
    a SqlAlchemy class model.
    """
    #: If Flask-SqlAlchemy is used then create an instance of the
    #: SqlAlchemy class and then pass this instance to the Column
    #: constructor::
    #:
    #:      from flask_sqlalchemy import SQLAlchemy
    #:
    #:      db = SQLAlchemy()
    #:      my_video = file_upload.Column(db)
    db = None

    def __init__(self, db=None):
        self.db = db



"""
    Helper class
"""
import os

from ._config import Config
from ._model_utils import _ModelUtils


class FileUtils:

    table_name: str

    model = None

    config: Config

    id: int

    def __init__(self, model, config: Config):
        self.config = config
        self.model = model
        self.id = _ModelUtils.get_primary_key(model)
        self.table_name = _ModelUtils.get_table_name(model)

    @staticmethod
    def allowed_file(filename, config: Config) -> bool:
        """
        :param filename:
        :param config:
        :return bool:
        """
        return "." in filename and \
            filename.rsplit(".", 1)[1].lower() in config.allowed_extensions

    def postfix_file_path(self, id: int, filename: str) -> str:
        """
        :param id:
        :param filename:
        :return str:
        """
        return f"/{self.table_name}/{id}/{filename}"

    def get_file_path(self, model_id: int, filename: str) -> str:
        """
        :param model_id:
        :param filename:
        :return str:
        """
        return os.path.join(f"{self.config.upload_folder}{self.postfix_file_path(model_id, filename)}")

    def save_file(self, file, model_id: int) -> None:
        """
        :param file:
        :param model_id:
        :return None:
        """
        file.save(self.get_file_path(model_id, file.filename))

    def get_stream_path(self, model_id: int):
        return os.path.join(f"{self.config.upload_folder}/{self.table_name}/{model_id}")
